# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the ForbiddenRealComparisons rule."""

from typing import Dict, Optional, Set, Tuple

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.predef import (
    SC_ECK_BLD_STRUCT,
    SC_ECK_CASE,
    SC_ECK_DIV,
    SC_ECK_EQUAL,
    SC_ECK_FBY,
    SC_ECK_FOLLOW,
    SC_ECK_GEQUAL,
    SC_ECK_GREAT,
    SC_ECK_IF,
    SC_ECK_INT2REAL,
    SC_ECK_LEQUAL,
    SC_ECK_LESS,
    SC_ECK_MAKE,
    SC_ECK_NEQUAL,
    SC_ECK_NUMERIC_CAST,
    SC_ECK_PRE,
    SC_ECK_PRJ,
    SC_ECK_PRJ_DYN,
    SC_ECK_REAL2INT,
    SC_ECK_SEQ_EXPR,
    SC_ECK_TIMES,
)
from ansys.scade.design_rules.utils.rule import Rule


class Mask:
    """
    Base class for abstracting a type.

    A mask is considered as float if one of its components is float.
    """

    def __init__(self, is_float=None):
        # cache
        self._is_float = is_float

    def is_float(self) -> bool:
        """Return whether the type is float."""
        return self._is_float is True


class Scalar(Mask):
    """Mask class for scalar types or arrays."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Struct(Mask):
    """Mask class for structure types."""

    def __init__(self, fields=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields = fields if fields else {}

    def is_float(self) -> bool:
        """Return whether the type is float."""
        if self._is_float is None:
            # lazy evaluation
            self._is_float = False
            for _, mask in self.fields.items():
                self._is_float |= mask is not None and mask.is_float()
        return self._is_float


class ForbiddenRealComparisons(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0024',
        description='Comparisons of real values are not allowed: = (20, 21, 22, 23, 24, 25)',
        label='Forbidden real comparisons',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='20, 21, 22, 23, 24, 25',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.ExprCall],
            kinds=None,
        )
        # comparison operators specified in the rule's parameters
        self.comp_operators = set()
        # predefined float types
        self.floats: Set[suite.Type] = set()
        # cache for type abstractions
        self.masks: Dict[Optional[suite.Type], Optional[Mask]] = {}

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Cache the predefined types and get the parameters."""
        if not parameter:
            self.set_message('No parameter given, rule cannot be checked.')
            return Rule.ERROR

        self.comp_operators = {int(_) for _ in parameter.split(',') if _} if parameter else set()
        # float types for Scade 6.4 or 6.6
        self.floats = set()
        for name in ('float', 'float32', 'float64'):
            type_ = model.session.find_predefined_type(name)
            if type_:
                self.floats.add(type_)
        # reset cache
        self.masks = {}

        return Rule.OK

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        assert isinstance(object_, suite.ExprCall)  # nosec B101  # addresses linter
        if object_.predef_opr in self.comp_operators:
            real_found, error = self.handle_expressions(object_.parameters)

            if error:
                # ERROR leads to a fatal error, not correctly reported
                # self.set_message("Error evaluating " + object.to_string() + ", ")
                # return Rule.ERROR
                Rule.OK
            elif real_found:
                self.set_message(f'Real comparison in {object_.to_string()}, ')
                return Rule.FAILED

        return Rule.OK

    def handle_expressions(self, expressions) -> Tuple[bool, bool]:
        """Recurse on expressions elements."""
        result = False
        error = False
        for expr in expressions:
            mask = self.get_expression_mask(expr)
            if mask is None:
                error = True
            else:
                result |= not error and mask.is_float()
        return result, error

    def get_expression_mask(self, expr: suite.Expression) -> Optional[Mask]:
        """Return the type mask for expr."""
        default = Scalar(is_float=False)
        if isinstance(expr, suite.ConstValue):
            # if not float, the returned type does not matter
            return Scalar(is_float=expr.kind == 'Float')
        elif isinstance(expr, suite.ExprId):
            # expr.reference might be None with incomplete models
            return self.get_type_mask(expr.reference.type) if expr.reference else None
        elif isinstance(expr, suite.ExprType):
            # ExprType can't be an operand of a comparison
            # must not be called
            pass  # pragma: no cover
        elif isinstance(expr, suite.ExprText):
            # syntax error
            # must not be called
            pass  # pragma: no cover

        assert isinstance(expr, suite.ExprCall)  # nosec B101  # addresses linter
        if expr.operator:
            # must be a function, check the type ouf the first output
            outputs = expr.operator.outputs
            # assert len(outputs) == 1
            return self.get_type_mask(outputs[0].type) if outputs else default

        # predefined operator: consider the parameters
        # for most of the operators, it is enough to consider the first parameter
        code = expr.predef_opr
        parameters = expr.parameters
        if code in {
            SC_ECK_LESS,
            SC_ECK_LEQUAL,
            SC_ECK_GREAT,
            SC_ECK_GEQUAL,
            SC_ECK_EQUAL,
            SC_ECK_NEQUAL,
            SC_ECK_TIMES,
        }:
            # boolean operator
            return default
        elif code in {SC_ECK_DIV, SC_ECK_REAL2INT}:
            # ensure compatibility with Scade 6.4
            return default
        elif code in {SC_ECK_INT2REAL}:
            # ensure compatibility with Scade 6.4
            return Scalar(is_float=True)
        elif code in {SC_ECK_FOLLOW}:
            # type of the first parameter of the first sequence
            return self.get_expression_mask(expr.parameters[0].parameters[0])
        elif code in {SC_ECK_PRE, SC_ECK_FBY}:
            # type of the first parameter
            return self.get_expression_mask(expr.parameters[0])
        elif code in {SC_ECK_IF, SC_ECK_CASE}:
            # type of the first 'then' or 'case' parameter
            return self.get_expression_mask(expr.parameters[1].parameters[0])
        elif code == SC_ECK_SEQ_EXPR:
            # must not happen
            raise AssertionError('internal error: SEQ_EXPR')  # pragma: no cover
        elif code in {SC_ECK_NUMERIC_CAST, SC_ECK_MAKE}:
            assert isinstance(expr.parameters[1], suite.ExprType)  # nosec B101  # addresses linter
            return self.get_type_mask(expr.parameters[1].reference)
        elif code == SC_ECK_BLD_STRUCT:
            fields = {_.label.name: self.get_expression_mask(_) for _ in parameters}
            return Struct(fields)
        elif code == SC_ECK_PRJ:
            mask = self.get_expression_mask(parameters[0])
            for filter in parameters[1:]:
                if isinstance(filter, suite.ConstValue) and filter.kind == 'Label':
                    if not isinstance(mask, Struct):
                        # semantic error
                        return None
                    mask = mask.fields.get(filter.value)
                # else: ignore indexes
            return mask
        elif code == SC_ECK_PRJ_DYN:
            # consider the default value
            return self.get_expression_mask(parameters[-1])
        else:
            return self.get_expression_mask(parameters[0])

    def get_type_mask(self, type_: suite.Type) -> Optional[Mask]:
        """Return the mask of a type."""
        while isinstance(type_, suite.NamedType):
            if type_.is_predefined() or type_.is_generic():
                break
            type_ = type_.type
        mask = self.masks.get(type_)
        if not mask:
            if isinstance(type_, suite.NamedType):
                if type_.constraint:
                    # polymorphic type
                    is_float = type_.constraint.is_float()
                elif type_.is_predefined():
                    is_float = type_ in self.floats
                else:
                    # incomplete model, alias without definition
                    # or unconstrained polymorphic type
                    is_float = False
                mask = Scalar(is_float)
            elif isinstance(type_, suite.Table):
                return self.get_type_mask(type_.type)
            elif isinstance(type_, suite.Structure):
                fields = {_.name: self.get_type_mask(_.type) for _ in type_.elements}
                mask = Struct(fields)
            elif type_:
                # enumerations, sized types
                mask = Scalar(is_float=False)
            self.masks[type_] = mask
        return mask


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    ForbiddenRealComparisons()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the ElementsWithinArea rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import re

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class ElementsWithinArea(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0015',
        description=('All elements shall be within the page format (A4, A3).'),
        label='Elements within area.',
        category='Diagrams',
        parameter='margins=0; 0',
        severity=Rule.REQUIRED,
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[
                suite.Flow,
                suite.Transition,
            ],
            kinds=None,
        )
        # parameters
        self.x_margin = 0
        self.y_margin = 0
        # cache for boundaries
        self.cached_diagram = None
        self.cached_status = False
        # current format
        self.format = None
        # current left boundary
        self.min_x = 0
        # current upper boundary
        self.min_y = 0
        # current right boundary
        self.max_x = 0
        # current lower boundary
        self.max_y = 0
        # cache to report wrong formats only once
        self.wrong_formats = set()

        # formats available in the SCADE Editor, 1/100th of mm
        self.boundaries = {
            'A3': [29700, 42000],
            'A4': [21000, 29700],
            'B5': [18200, 25700],
            'LETTER': [21590, 27940],
            'LEGAL': [21590, 35560],
        }

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Reset the caches and parse the parameters."""
        self.wrong_formats = set()

        # default message if syntax error
        message = f"'{parameter}': syntax error, expected 'margins= <value>; <value>'"
        parameter = 'margins= 0; 0' if not parameter else parameter
        d = self.parse_values(parameter)
        if d is not None:
            margins = d.get('margins', '')
            try:
                self.x_margin, self.y_margin = (int(_) for _ in margins.split(';'))
                return Rule.OK
            except BaseException:
                pass

        self.set_message(message)
        return Rule.ERROR

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        pe = object_.presentation_element
        if not pe or not isinstance(pe.diagram, suite.NetDiagram):
            return Rule.NA

        diagram = pe.diagram
        status = self.set_boundaries(diagram)
        if not status:
            if diagram not in self.wrong_formats:
                # cache the diagram and not the format to report
                # the error once per diagram
                self.wrong_formats.add(diagram)
                self.set_message(f'Format {self.format} not supported')
                return Rule.FAILED
            else:
                return Rule.NA

        items = set()
        if isinstance(pe, suite.Box):
            status = not self.is_box_outside_area(pe.position, pe.size)
            if isinstance(pe, suite.EquationGE):
                if not status:
                    items.add('equation')
                # check outgoing edges
                for edge in pe.out_edges:
                    if self.is_link_outside_area(edge.points):
                        # do not use logical shortcut to report all findings
                        status = False
                        items.add(f'edge {edge.left_var.name}')
        else:
            assert isinstance(pe, suite.TransitionGE)  # nosec B101  # addresses linter
            if self.is_box_outside_area(pe.label_pos, pe.label_size):
                status = False
                items.add('label')
            if self.is_link_outside_area(pe.points):
                status = False
                items.add('link')

        if not status:
            message = 'Element outside area'
            if items:
                message += ' ({})'.format(', '.join(sorted(items)))

            self.set_message(message)
            return Rule.FAILED

        return Rule.OK

    def set_boundaries(self, diagram: suite.NetDiagram) -> bool:
        """Cache the boundaries and margins for the input diagram."""
        if diagram == self.cached_diagram:
            return self.cached_status
        self.cached_diagram = diagram
        self.format = diagram.format.split()[0]
        if self.format not in self.boundaries.keys():
            # custom format
            regexp = r'(.*)\s*\(\s*(\d+)\s+(\d+)\s*\)\s*'
            match = re.match(regexp, diagram.format)
            if match:
                groups = match.groups()
                self.format = groups[0]
                self.boundaries[self.format] = [int(_) * 100 for _ in groups[1:]]
            else:
                self.cached_status = False
                return False

        if diagram.landscape:
            self.max_x = self.boundaries[self.format][1] - self.y_margin
            self.max_y = self.boundaries[self.format][0] - self.x_margin
        else:
            self.max_x = self.boundaries[self.format][0] - self.x_margin
            self.max_y = self.boundaries[self.format][1] - self.y_margin

        self.cached_status = True
        return True

    def is_box_outside_area(self, pos, size) -> bool:
        """Return whether an box is within the area."""
        x, y = pos
        w, h = size
        return x < self.min_x or y < self.min_y or x + w > self.max_x or y + h > self.max_y

    def is_link_outside_area(self, points) -> bool:
        """Return whether a line is within the area."""
        for x, y in points:
            if x < self.min_x or y < self.min_y or x > self.max_x or y > self.max_y:
                return True
        return False


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    ElementsWithinArea()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the PageFormat rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class PageFormat(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0113',
        category='Diagrams',
        severity=Rule.REQUIRED,
        parameter='format=A4,orientation=Landscape',
        label='Check page format.',
        types=None,
        description=(
            'This rule checks if the page format is set properly.\n'
            'param1: format=A3 A4 B5 etc.\n'
            'param2: orientation=Portrait Landscape\n'
            'For example: format=A3,orientation=Landscape'
        ),
    ):
        if not types:
            types = [suite.NetDiagram]
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=types,
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            d = f"'{parameter}': parameter syntax error"
        else:
            format = d.get('format')
            orientation = d.get('orientation')
            if format is None:
                message = f"'{parameter}': missing 'format' value"
            elif orientation is None:
                message = f"'{parameter}': missing 'orientation' value"
            else:
                self.format = format
                self.orientation = orientation
                return Rule.OK

        self.set_message(message)
        return Rule.ERROR

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        violated = False
        failure_messages = []

        net_diagram = object

        if not net_diagram.format.startswith(self.format):
            violated = True
            failure_messages.append(f'Format is set to {net_diagram.format}')

        if net_diagram.landscape and self.orientation != 'Landscape':
            violated = True
            failure_messages.append('Orientation is set to Landscape.')

        if not net_diagram.landscape and self.orientation != 'Portrait':
            violated = True
            failure_messages.append('Orientation is set to Portrait.')

        if violated:
            self.set_message(','.join(failure_messages))
            return Rule.FAILED

        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    PageFormat()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides a common base class for rules, with additional services.

This class stubs the original ``Rule`` class for unit tests.
"""

from typing import Dict, List, Optional

import scade.model.suite as suite

try:
    # ignore F401: _register_rule made available for rules, not used here
    # _register_rule is defined dynamically: ignore linter warning
    from scade.tool.suite import _register_rule  # noqa: F401  # type: ignore

    # _Rule defined hereafter is a stub, and thus can't match Rule
    from scade.tool.suite.rules import Rule as _Rule  # type: ignore
except ImportError:
    from .metric import Metric

    class _Rule:
        """Stubs ``Rule`` for units tests."""

        def __init__(self, *args, **kwargs):
            self.message = ''
            self.metrics = {}

        def set_message(self, message: str) -> None:
            """Store the message."""
            self.message = message

        def add_rule_status(
            self,
            object_: suite.Object,
            status: int,
            message: str,
            local_id: str = '',
        ) -> None:
            """Stub the ``add_rule_status`` method."""
            pass

        def stub_metrics(self, metrics: Dict[str, Metric]):
            """
            Store the metric instances for the evaluation.

            This method allows emulating the environment for unit tests.
            """
            self.metrics = metrics

        def get_metric_result(self, object_: suite.Object, metric_id: str) -> int:
            """
            Compute the value of a metric for an object.

            This method allows emulating the environment for unit tests.
            """
            metric = self.metrics[metric_id]
            status = metric.on_compute(object_)
            assert status == Metric.OK  # nosec B101  # used for unit tests
            return metric.result

        # Possible severity values
        MANDATORY = 0
        REQUIRED = 1
        ADVISORY = 2

        # Possible returned values by the following functions
        OK = 0
        FAILED = 1
        ERROR = 2
        # new 2023 R1, stands for NA
        NO_RULE = 3


from .sck import SCK


class Rule(_Rule):
    """Base class for all rules."""

    # set NA to OK until it is implemented
    # new 2023 R1: NA is there but named NO_RULE: let's keep NA
    try:
        NA = _Rule.NO_RULE
    except BaseException:
        NA = 0

    def __init__(
        self,
        id: str,
        label: str,
        category: str,
        severity: int,
        types: Optional[List[type]] = None,
        kinds: Optional[List[SCK]] = None,
        **kwargs,
    ):
        # discrimination of SCADE objects when the class isn't enough
        # compute the list of types when kinds is specified
        if kinds:
            types = [kind.value[0] for kind in kinds]
            self.kinds = kinds
        else:
            self.kinds = None
        # store the types, useful for unit tests
        self.types = types
        super().__init__(id, label, category, severity, types, **kwargs)
        self.violations = {}

    def on_stop(self) -> int:
        """Flush the violations."""
        # 2024 R2: crash when calling add_rule_status from on_stop
        # for (object_, local_id), (status, message) in self.violations.items():
        #     try:
        #         # 2024 R2 and above
        #         super().add_rule_status(object_, status, message, local_id)
        #     except TypeError:
        #         super().add_rule_status(object_, status, message)
        self.violations = {}
        return Rule.OK

    def add_rule_status(
        self,
        object_: suite.Object,
        status: int,
        message: str,
        local_id: str = '',
    ) -> None:
        """Register the violation."""
        key = (object_, local_id)
        # 2024 R2: crash when calling add_rule_status from on_stop
        # --> call add_rule_status right now if not already present
        # self.violations[key] = (status, message)
        if key not in self.violations:
            self.violations[key] = (status, message)
            try:
                # 2024 R2 and above
                super().add_rule_status(object_, status, message, local_id)
            except TypeError:
                super().add_rule_status(object_, status, message)

    def get_closest_annotatable(self, object_: suite.Object) -> suite.Object:
        """Return the closest container that can be annotated; Can be the object itself."""
        if isinstance(object_, suite.Edge):
            object_ = object_.src_equation
        if isinstance(object_, suite.PresentationElement):
            object_ = object_.presentable
        while not isinstance(object_, suite.Annotable):
            object_ = object_.owner
        # exceptions
        if isinstance(object_, suite.NamedType):
            if object_.is_generic():
                # generic types do not have an oid: return the operator
                return object_.owner
        elif isinstance(object_, suite.Type):
            # only named types can be annotated: recurse
            return self.get_closest_annotatable(object_.owner)
        return object_

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Apply the rule to the input if it is compatible with the registered kinds."""
        if self.accept_kind(object_):
            return self.on_check_ex(object_, parameter)
        else:
            return Rule.NA

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Apply the rule to the input object."""
        # must not be called if on_check is redefined
        return Rule.ERROR

    def accept_kind(self, object_: suite.Object):
        """Return whether the input is compatible with the registered kinds."""
        if not self.kinds:
            return True

        for kind in self.kinds:
            type, filter = kind.value
            if isinstance(object_, type) and filter(object_):
                return True
        return False

    def parse_values(self, parameter: str):
        """Compile a string containing comma separated values into a dictionary."""
        d = {}
        for values in parameter.split(','):
            try:
                name, value = values.split('=')
                d[name.strip()] = value.strip()
            except BaseException:
                return None
        return d



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the PragmaManifest rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade
import scade.model.suite as suite

import ansys.scade.apitools.prop as prop
import ansys.scade.apitools.query as query
from ansys.scade.design_rules.utils.rule import SCK, Rule


class PragmaManifest(Rule):
    """Implements the rule interface."""

    # abbreviations for computing the signatures of the types
    _predef_signatures = {
        'bool': 'b',
        'char': 'c',
        'int8': 'i8',
        'int16': 'i16',
        'int32': 'i32',
        'int64': 'i64',
        'uint8': 'u8',
        'uint16': 'u16',
        'uint32': 'u32',
        'uint64': 'u64',
        'float32': 'f32',
        'float64': 'f64',
    }

    def __init__(
        self,
        id='id_0123',
        label='Pragma manifest for complex types',
        description=(
            "A pragma 'manifest' shall be used for each type declaration.\n"
            'Optionally, the rule applies only for each type declaration '
            'used in the interface of a root operator / imported operator: '
            'Either all the root operators, or the ones of '
            'the specified configuration.'
        ),
        category='Modelling',
        severity=Rule.ADVISORY,
        parameter='configuration=,interface=false',
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            kinds=[SCK.TYPE],
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )
        self.roots = None

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        # restore the default values
        self.roots = None
        # the parameters are optional
        d = self.parse_values(parameter) if parameter else {}
        if d is None:
            message = f"'{parameter}': parameter syntax error"
            self.set_message(message)
            # scade is a CPython module defined dynamically
            scade.output(message + '\n')  # type: ignore
            return Rule.ERROR
        name = d.get('configuration')
        self.interface = d.get('interface', 'false') == 'true'
        if name:
            # get the project associated to the model
            project = model.project
            configuration = project.find_configuration(name)
            if configuration:
                self.roots = project.get_tool_prop_def('GENERATOR', 'ROOTNODE', [], configuration)

        # cache dictionary type -> signature
        self.signatures = {}
        # cache dictionary signature -> has pragma manifest
        self.manifests = {}
        # store the presence of a pragma manifest for all type equivalence classes
        # from the model and its libraries.
        for type_ in model.all_named_types:
            # consider only structures and arrays
            if not (query.is_structure(type_) or query.is_array(type_)):
                continue
            self._update_manifest(type_)

        # no error
        return Rule.OK

    def on_check_ex(self, type_: suite.NamedType, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if not (query.is_structure(type_) or query.is_array(type_)):
            # the pragma manifest is not required for imported or scalar types
            return Rule.NA
        if self.interface:
            # search the root operators where the type is used, as main type, in the interface
            operators = set()
            for typed in type_.typed_objects:
                if isinstance(typed, suite.LocalVariable) and (
                    typed.is_input() or typed.is_hidden() or typed.is_output()
                ):
                    if self._is_root(typed.operator):
                        operators.add(typed.operator.get_full_path().strip('/'))
            if not operators:
                # the type is not directly involved in the signature of a root operator
                return Rule.NA
            # consider the type and its dependencies
            # get the equivalence class using the signature
            for used in type_.used_types:
                if not (query.is_structure(used) or query.is_array(used)):
                    continue
                signature = self.get_signature(used)
                if not self.manifests[signature]:
                    message = (
                        'The type {} has no KCG pragma "manifest" and is'
                        ' used in the interface of root operators: {}'
                    )
                    self.set_message(message.format(query.get_type_name(used), operators))
                    return Rule.FAILED
        else:
            signature = self.get_signature(type_)
            if not self.manifests[signature]:
                self.set_message(
                    f'The type {query.get_type_name(type_)} has no KCG pragma "manifest"'
                )
                return Rule.FAILED
        return Rule.OK

    def _is_root(self, operator: suite.Operator) -> bool:
        # no configuration specified, consider any root operator
        # which is not polymorphic nor parameterized by size
        if operator.parameters or operator.typevars:
            return False
        return (
            operator.is_imported()
            or not operator.expr_calls
            # configuration specified: consider its root operators
            and (not self.roots or operator.get_full_path().strip('/') in self.roots)
        )

    def get_signature(self, type_) -> str:
        """Return the signature of the type."""
        signature = self.signatures.get(type_)
        if not signature:
            if isinstance(type_, suite.NamedType):
                if type_.is_imported():
                    signature = '({})'.format(type_.name)
                elif type_.is_predefined():
                    signature = self._predef_signatures[type_.name]
                else:
                    signature = self.get_signature(type_.type)
            elif isinstance(type_, suite.Enumeration):
                signature = '({})'.format(type_.owner.name)
            elif isinstance(type_, suite.Table):
                signature = '{}^{:.0f}'.format(self.get_signature(type_.type), type_.size)
            elif isinstance(type_, suite.Structure):
                elements = [f'{_.name}:{self.get_signature(_.type)}' for _ in type_.elements]
                signature = '({})'.format(','.join(elements))
            elif isinstance(type_, suite.SizedType):
                prefix = 'u' if type_.constraint.is_unsigned() else 'i'
                size = type_.size_expression.evaluate_expression()
                if not size:
                    size = '({})'.format(type_.size_expression.to_string())
                signature = prefix + size
            else:
                # assert type_ is None
                signature = '()'
            self.signatures[type_] = signature

        return signature

    def _update_manifest(self, type_):
        manifest = prop.find_pragma_tool(type_, 'kcg', 'manifest') is not None
        # update the status for the type's equivalence class
        signature = self.get_signature(type_)
        self.manifests[signature] = manifest or self.manifests.setdefault(signature, False)


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    PragmaManifest()
